"""contains model classes representing MIGs and AHBs"""

from .anwendungshandbuch import Anwendungshandbuch
from .messageimplementationguide import MessageImplementationGuide

__all__ = ["MessageImplementationGuide", "Anwendungshandbuch"]
